from edu_rdm_integration.collect_and_export_data.models import (
    EduRdmCollectDataCommandProgress,
    EduRdmExportDataCommandProgress,
)
from edu_rdm_integration.collect_data.collect import (
    BaseCollectModelsDataByGeneratingLogs,
)
from edu_rdm_integration.export_data.export import (
    ExportEntitiesData,
)
from edu_rdm_integration.helpers import (
    save_command_log_link,
)


class CollectCommandMixin:
    """Класс-примесь для запуска команды сборки моделей."""

    def get_collect_command(self, command_id: int) -> EduRdmCollectDataCommandProgress:
        """Возвращает экземпляр модели команды запуска."""
        command = EduRdmCollectDataCommandProgress.objects.get(id=command_id)

        return command

    def get_collect_models_class(self):
        """Возвращает класс для сбора данных."""
        return BaseCollectModelsDataByGeneratingLogs

    def run_collect_command(self, command) -> None:
        """Запуск команды сбора."""
        collect_models_data_class = self.get_collect_models_class()
        collect_models_data_class(
            models=(command.model_id,),
            logs_period_started_at=command.logs_period_started_at,
            logs_period_ended_at=command.logs_period_ended_at,
            command_id=command.id,
            institute_ids=tuple(command.institute_ids or ()),
        ).collect()

    def save_collect_command_logs(self, command_id: int, log_dir: str):
        """Сохранение ссылки на файл логов в команде."""
        try:
            command = self.get_collect_command(command_id)
        except EduRdmCollectDataCommandProgress.DoesNotExist:
            command = None

        if command:
            save_command_log_link(command, log_dir)


class ExportCommandMixin:
    """Класс-примесь для запуска команды выгрузки сущностей."""

    def get_export_command(self, command_id: int) -> EduRdmExportDataCommandProgress:
        """Возвращает экземпляр модели команды запуска."""
        command = EduRdmExportDataCommandProgress.objects.get(id=command_id)

        return command

    def run_export_command(self, command: EduRdmExportDataCommandProgress) -> None:
        """Запуск команды выгрузки."""
        ExportEntitiesData(
            entities=(command.entity_id,),
            period_started_at=command.period_started_at,
            period_ended_at=command.period_ended_at,
            command_id=command.id,
        ).export()

    def save_export_command_logs(self, command_id: int, log_dir: str):
        """Сохранение ссылки на файл логов в команде."""
        try:
            command = self.get_export_command(command_id)
        except EduRdmExportDataCommandProgress.DoesNotExist:
            command = None

        if command:
            save_command_log_link(command, log_dir)



from abc import (
    ABCMeta,
)
from typing import (
    Type,
)

from edu_rdm_integration.collect_data.base.functions import (
    BaseCollectingCalculatedDataFunction,
)
from edu_rdm_integration.collect_data.non_calculated.base.helpers import (
    BaseCollectingExportedDataFunctionHelper,
)
from edu_rdm_integration.collect_data.non_calculated.base.results import (
    BaseCollectingExportedDataFunctionResult,
)
from edu_rdm_integration.collect_data.non_calculated.base.validators import (
    BaseCollectingExportedDataFunctionValidator,
)


class BaseCollectingExportedDataFunction(BaseCollectingCalculatedDataFunction, metaclass=ABCMeta):
    """
    Базовый класс функций сбора данных для интеграции с "Региональная витрина данных".
    """

    def _prepare_helper_class(self) -> Type[BaseCollectingExportedDataFunctionHelper]:
        """
        Возвращает класс помощника функции.
        """
        return BaseCollectingExportedDataFunctionHelper

    def _prepare_validator_class(self) -> Type[BaseCollectingExportedDataFunctionValidator]:
        """
        Возвращает класс валидатора функции.
        """
        return BaseCollectingExportedDataFunctionValidator

    def _prepare_result_class(self) -> Type[BaseCollectingExportedDataFunctionResult]:
        """
        Возвращает класс результата функции.
        """
        return BaseCollectingExportedDataFunctionResult



from typing import (
    Type,
)

from edu_rdm_integration.adapters.runners import (
    WebEduRunner,
)
from edu_rdm_integration.export_data.base.helpers import (
    BaseExportDataRunnerHelper,
)
from edu_rdm_integration.export_data.base.results import (
    BaseExportDataRunnerResult,
)
from edu_rdm_integration.export_data.base.validators import (
    BaseExportDataRunnerValidator,
)


class BaseExportDataRunner(WebEduRunner):
    """
    Базовый класс ранеров функций выгрузки данных для интеграции с "Региональная витрина данных".
    """

    def _prepare_helper_class(self) -> Type[BaseExportDataRunnerHelper]:
        """
        Возвращает класс помощника ранера функции.
        """
        return BaseExportDataRunnerHelper

    def _prepare_validator_class(self) -> Type[BaseExportDataRunnerValidator]:
        """
        Возвращает класс валидатора ранера функции.
        """
        return BaseExportDataRunnerValidator

    def _prepare_result_class(self) -> Type[BaseExportDataRunnerResult]:
        """
        Возвращает класс результата ранера функции.
        """
        return BaseExportDataRunnerResult

    def _prepare_model_ids_chunks(self, *args, model_ids_map=None, **kwargs):
        """
        Формирование чанков идентификаторов записей моделей для дальнейшей работы в рамках функций.
        """
        # model_ids_chunks = make_chunks(
        #     iterable=model_ids,
        #     size=settings.RDM_EXPORT_CHUNK_SIZE,
        # )

        return ()

    def _get_runnable_objects(self, *args, **kwargs):
        """
        Возвращает генератор запускаемых объектов.
        """
        model_ids_chunks = self._prepare_model_ids_chunks(*args, **kwargs)

        for chunk_index, model_ids_chunk in enumerate(model_ids_chunks, start=1):
            for runnable_class in self._prepare_runnable_classes():
                runnable = runnable_class(
                    model_ids=list(model_ids_chunk),
                    chunk_index=chunk_index,
                    *args,
                    **kwargs,
                )

                if isinstance(runnable, tuple):
                    yield from runnable
                else:
                    yield runnable



# Generated by Django 2.2.28 on 2023-12-05 11:46
from django.db import (
    migrations,
)

from m3_db_utils.consts import (
    DEFAULT_ORDER_NUMBER,
)


def init_request_statuses(apps, schema_editor):
    """Заполнение первоначальными данными DataMartRequestStatus."""
    DataMartRequestStatus = apps.get_model('edu_rdm_integration', 'DataMartRequestStatus')  # noqa: N806

    DataMartRequestStatus.objects.bulk_create([
        DataMartRequestStatus(
            key='UPLOAD_TO_BUFFER',
            value=-1,
            title='Загрузка данных в буффер',
            order_number=DEFAULT_ORDER_NUMBER,
        ),
        DataMartRequestStatus(
            key='BUFFERED',
            value=0,
            title='Запрос буфферизирован',
            order_number=DEFAULT_ORDER_NUMBER,
        ),
        DataMartRequestStatus(
            key='WAIT_FOR_OPEN_DELTA',
            value=1,
            title='Ожидает открытия дельты',
            order_number=DEFAULT_ORDER_NUMBER,
        ),
        DataMartRequestStatus(
            key='IN_PROCESSING',
            value=2,
            title='В обработке',
            order_number=DEFAULT_ORDER_NUMBER,
        ),
        DataMartRequestStatus(
            key='SUCCESSFULLY_PROCESSED',
            value=3,
            title='Успешно обработан',
            order_number=DEFAULT_ORDER_NUMBER,
        ),
        DataMartRequestStatus(
            key='FAILED_PROCESSING',
            value=4,
            title='Ошибка обработки запроса',
            order_number=DEFAULT_ORDER_NUMBER,
        ),
        DataMartRequestStatus(
            key='REQUEST_ID_NOT_FOUND',
            value=5,
            title='Идентификатор запроса не обнаружен',
            order_number=DEFAULT_ORDER_NUMBER,
        ),
        DataMartRequestStatus(
            key='FORMAT_LOGICAL_CONTROL',
            value=6,
            title='Форматно-логический контроль',
            order_number=DEFAULT_ORDER_NUMBER,
        ),
        DataMartRequestStatus(
            key='FLC_ERROR',
            value=7,
            title='Ошибки ФЛК',
            order_number=DEFAULT_ORDER_NUMBER,
        ),
    ], ignore_conflicts=True)


def migrate_upload_status_to_request_status(apps, schema_editor):
    """Переносит значения из поля status в request_status модели UploadStatusRequestLog."""
    UploadStatusRequestLog = apps.get_model('edu_rdm_integration', 'UploadStatusRequestLog')  # noqa: N806
    DataMartRequestStatus = apps.get_model('edu_rdm_integration', 'DataMartRequestStatus')  # noqa: N806

    request_status_value_to_enum_key = dict(DataMartRequestStatus.objects.values_list('value', 'key'))

    for value, enum_key in request_status_value_to_enum_key.items():
        UploadStatusRequestLog.objects.filter(
            status__code=value,
        ).update(
            request_status=enum_key,
        )


class Migration(migrations.Migration):

    dependencies = [
        ('edu_rdm_integration', '0005_auto_20231204_1224'),
    ]

    operations = [
        migrations.RunPython(init_request_statuses, reverse_code=migrations.RunPython.noop),
        migrations.RunPython(migrate_upload_status_to_request_status, reverse_code=migrations.RunPython.noop)
    ]



# Generated by Django 2.2.28 on 2023-12-05 14:04
# flake8: noqa
from django.db import (
    migrations,
)


class Migration(migrations.Migration):

    dependencies = [
        ('edu_rdm_integration', '0006_request_status_data'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='uploadstatusrequestlog',
            name='status',
        ),
        migrations.DeleteModel(
            name='UploadStatus',
        ),
    ]



# Generated by Django 3.0.14 on 2024-03-11 16:10
# flake8: noqa
import uuid

import django.contrib.postgres.fields.jsonb
import django.db.models.deletion
import django.utils.timezone
from django.db import (
    migrations,
    models,
)

import educommon.django.db.mixins

import edu_rdm_integration.utils


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('edu_rdm_integration', '0008_transferredentity'),
        ('async_task', '0002_task_type_and_status_data'),
    ]

    operations = [
        migrations.CreateModel(
            name='EduRdmExportDataCommandProgress',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(default=django.utils.timezone.now, verbose_name='Дата создания')),
                ('period_started_at', models.DateTimeField(verbose_name='Левая граница периода выборки данных для выгрузки')),
                ('period_ended_at', models.DateTimeField(verbose_name='Правая граница периода выборки данных для выгрузки')),
                ('generation_id', models.UUIDField(default=uuid.uuid4, verbose_name='Идентификатор генерации')),
                ('logs_link', models.FileField(max_length=255, upload_to=edu_rdm_integration.utils.get_data_command_progress_attachment_path, verbose_name='Файл лога')),
                ('type', models.PositiveSmallIntegerField(choices=[(1, 'Автоматический'), (2, 'Ручной')], verbose_name='Тип команды')),
                ('entity', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='edu_rdm_integration.RegionalDataMartEntityEnum', verbose_name='Сущность РВД')),
                ('stage', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='edu_rdm_integration.ExportingDataStage', verbose_name='Этап выгрузки данных')),
                ('task', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='async_task.RunningTask', verbose_name='Асинхронная задача')),
            ],
            options={
                'db_table': 'edu_rdm_exporting_data_command_progress',
            },
            bases=(educommon.django.db.mixins.ReprStrPreModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='EduRdmCollectDataCommandProgress',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(default=django.utils.timezone.now, verbose_name='Дата создания')),
                ('logs_period_started_at', models.DateTimeField(verbose_name='Левая граница периода обрабатываемых логов')),
                ('logs_period_ended_at', models.DateTimeField(verbose_name='Правая граница периода обрабатываемых логов')),
                ('generation_id', models.UUIDField(default=uuid.uuid4, verbose_name='Идентификатор генерации')),
                ('institute_ids', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=list, null=True, verbose_name='id учебного заведения')),
                ('logs_link', models.FileField(max_length=255, upload_to=edu_rdm_integration.utils.get_data_command_progress_attachment_path, verbose_name='Ссылка на файл логов')),
                ('type', models.PositiveSmallIntegerField(choices=[(1, 'Автоматический'), (2, 'Ручной')], verbose_name='Тип команды')),
                ('model', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='edu_rdm_integration.RegionalDataMartModelEnum', verbose_name='Модель РВД')),
                ('stage', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='edu_rdm_integration.CollectingExportedDataStage', verbose_name='Этап формирования данных для выгрузки')),
                ('task', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='async_task.RunningTask', verbose_name='Асинхронная задача')),
            ],
            options={
                'db_table': 'edu_rdm_collecting_data_command_progress',
            },
            bases=(educommon.django.db.mixins.ReprStrPreModelMixin, models.Model),
        ),
    ]



import json
from abc import (
    ABCMeta,
)
from pathlib import (
    Path,
)
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

from django.conf import (
    settings,
)
from django.core.files.base import (
    ContentFile,
)
from django.core.files.storage import (
    default_storage,
)
from transliterate import (
    slugify,
)
from uploader_client.adapters import (
    adapter,
)

from educommon import (
    logger,
)
from educommon.integration_entities.enums import (
    EntityLogOperation,
)
from educommon.integration_entities.mixins import (
    EntitiesMixin,
)

from edu_rdm_integration.adapters.functions import (
    WebEduLazySavingPredefinedQueueGlobalHelperFunction,
)
from edu_rdm_integration.consts import (
    LOGS_DELIMITER,
)
from edu_rdm_integration.enums import (
    FileUploadStatusEnum,
)
from edu_rdm_integration.export_data.base.consts import (
    OPERATIONS_METHODS_MAP,
    OPERATIONS_URLS_MAP,
)
from edu_rdm_integration.export_data.base.errors import (
    BaseExportDataError,
)
from edu_rdm_integration.export_data.base.helpers import (
    BaseExportDataFunctionHelper,
)
from edu_rdm_integration.export_data.base.requests import (
    RegionalDataMartEntityRequest,
)
from edu_rdm_integration.export_data.base.results import (
    BaseExportDataFunctionResult,
)
from edu_rdm_integration.export_data.base.validators import (
    BaseExportDataFunctionValidator,
)
from edu_rdm_integration.export_data.consts import (
    DELIMITER,
)
from edu_rdm_integration.models import (
    ExportingDataStage,
    ExportingDataSubStage,
    ExportingDataSubStageAttachment,
    ExportingDataSubStageStatus,
    ExportingDataSubStageUploaderClientLog,
)
from edu_rdm_integration.utils import (
    get_exporting_data_stage_attachment_path,
)


class BaseExportDataFunction(
    EntitiesMixin,
    WebEduLazySavingPredefinedQueueGlobalHelperFunction,
    metaclass=ABCMeta,
):
    """
    Базовый класс функций выгрузки данных для интеграции с "Региональная витрина данных".
    """

    def __init__(self, *args, stage: ExportingDataStage, model_ids: List[Union[int, str]], **kwargs):
        super().__init__(*args, entities=self.entities, model_ids=model_ids, **kwargs)

        self._sub_stage = ExportingDataSubStage.objects.create(
            stage=stage,
            function_id=self.uuid,
        )

        self._chunk_index = kwargs.get('chunk_index')

        logger.info(f'{LOGS_DELIMITER * 3}{repr(self._sub_stage)} created.')

        self._file_name = f'rdm_{self.first_entity.key.lower()}.csv'

        self._data = {
            EntityLogOperation.CREATE: [],
            EntityLogOperation.UPDATE: [],
            EntityLogOperation.DELETE: [],
        }

        self._file_contents: Dict[int, Optional[Tuple[Optional[str], Optional[ExportingDataSubStageAttachment]]]] = {
            EntityLogOperation.CREATE: (None, None),
            EntityLogOperation.UPDATE: (None, None),
            EntityLogOperation.DELETE: (None, None),
        }

    def _prepare_helper_class(self) -> Type[BaseExportDataFunctionHelper]:
        """
        Возвращает класс помощника функции.
        """
        return BaseExportDataFunctionHelper

    def _prepare_validator_class(self) -> Type[BaseExportDataFunctionValidator]:
        """
        Возвращает класс валидатора функции.
        """
        return BaseExportDataFunctionValidator

    def _prepare_result_class(self) -> Type[BaseExportDataFunctionResult]:
        """
        Возвращает класс результата функции.
        """
        return BaseExportDataFunctionResult

    def _before_prepare(self, *args, **kwargs):
        """
        Выполнение действий функций системы.
        """
        self._sub_stage.status_id = ExportingDataSubStageStatus.IN_PROGRESS.key
        self._sub_stage.save()

        logger.info(f'{LOGS_DELIMITER * 3}change status {repr(self._sub_stage)}')

    def _prepare_record(self, entity_instance) -> List[str]:
        """
        Формирование списка строковых значений полей.
        """
        ordered_fields = self.first_entity.entity.get_ordered_fields()
        primary_key_fields = set(self.first_entity.entity.get_primary_key_fields())
        foreign_key_fields = set(self.first_entity.entity.get_foreign_key_fields())
        required_fields = set(self.first_entity.entity.get_required_fields())
        hashable_fields = set(self.first_entity.entity.get_hashable_fields())
        ignore_prefix_fields = set(self.first_entity.entity.get_ignore_prefix_key_fields())

        field_values = self.helper.prepare_record(
            entity_instance=entity_instance,
            ordered_fields=ordered_fields,
            primary_key_fields=primary_key_fields,
            foreign_key_fields=foreign_key_fields,
            required_fields=required_fields,
            hashable_fields=hashable_fields,
            ignore_prefix_fields=ignore_prefix_fields,
        )

        return field_values

    def _prepare_data(self):
        """
        Преобразование собранных данных в удобный для выгрузки вид.
        """
        logger.info(f'{LOGS_DELIMITER * 3}{self.__class__.__name__} prepare data..')

        for entity_instance in self.helper.cache.entity_instances:
            self._data[entity_instance.operation].append(
                self._prepare_record(
                    entity_instance=entity_instance,
                )
            )

            entity_instance.exporting_sub_stage = self._sub_stage

        for operation in EntityLogOperation.values.keys():
            entities = self._data.get(operation)

            if entities:
                logger.info(
                    f'{LOGS_DELIMITER * 3}prepared {len(entities)} records with status '
                    f'{slugify(EntityLogOperation.values.get(operation))}..'
                )

    def _prepare_files(self):
        """
        Формирование файлов для дальнейшей выгрузки.
        """
        logger.info(f'{LOGS_DELIMITER * 3}{self.__class__.__name__} prepare files..')

        for operation in EntityLogOperation.values.keys():
            records = self._data[operation]
            if records:
                title_record = (
                    f'{DELIMITER}'.join([
                        field.lower()
                        for field in self.first_entity.entity.get_ordered_fields()
                    ])
                )

                joined_records = (
                    '\n'.join([
                        title_record,
                        *[f'{DELIMITER}'.join(record) for record in records]]
                    )
                )

                sub_stage_attachment = ExportingDataSubStageAttachment(
                    exporting_data_sub_stage=self._sub_stage,
                    operation=operation,
                )

                file_path = get_exporting_data_stage_attachment_path(
                    instance=sub_stage_attachment,
                    filename=self._file_name,
                )

                sub_stage_attachment.attachment = default_storage.save(file_path, ContentFile(joined_records))

                self.do_on_save(sub_stage_attachment)

                self._file_contents[operation] = (file_path, sub_stage_attachment)

    def _send_files(self):
        """
        Отправка файлов в "Региональная витрина данных".
        """
        # TODO При старте проекта и инициализации приложения function_tools происходит поиск стратегий. При поиске
        #  возникает ошибка "Не произведена настройка клиента". Настройка производится в конфиге
        #  приложения regional_data_mart_integration. Нужно доработать механизм так, чтобы сборка стратегий запускалась
        #  после готовности всех приложений.

        logger.info(f'{LOGS_DELIMITER * 3}{self.__class__.__name__} send files..')

        requests_result = []

        for operation, (relative_file_path, sub_stage_attachment) in self._file_contents.items():
            if relative_file_path:
                if settings.RDM_UPLOADER_CLIENT_ENABLE_REQUEST_EMULATION:
                    logger.warning(
                        f'{LOGS_DELIMITER * 3}ATTENTION!!! REGIONAL DATA MART INTEGRATION REQUEST EMULATION ENABLED!'
                    )

                method = OPERATIONS_METHODS_MAP.get(operation)
                file_path = Path.joinpath(Path(settings.MEDIA_ROOT), relative_file_path)

                request = RegionalDataMartEntityRequest(
                    datamart_name=settings.RDM_UPLOADER_CLIENT_DATAMART_NAME,
                    table_name=self.first_entity.key.lower(),
                    method=method,
                    operation=OPERATIONS_URLS_MAP.get(operation),
                    parameters={},
                    headers={
                        'Content-Type': 'text/csv',
                    },
                    files=[],
                    data=file_path.open('rb').read(),
                )

                result = adapter.send(request)

                request_id = ''
                if not result.error:
                    request_id = result.response.text

                file_upload_status = FileUploadStatusEnum.IN_PROGRESS if request_id else FileUploadStatusEnum.ERROR

                sub_stage_uploader_client_log = ExportingDataSubStageUploaderClientLog(
                    entry=result.log,
                    sub_stage=self._sub_stage,
                    attachment=sub_stage_attachment,
                    request_id=request_id,
                    file_upload_status=file_upload_status,
                    is_emulation=settings.RDM_UPLOADER_CLIENT_ENABLE_REQUEST_EMULATION,
                )

                self.do_on_save(sub_stage_uploader_client_log)

                if result.error:
                    logger.error(
                        f'{result.error}\nrequest - "{result.log.request}"\nresponse - "{result.log.response}"'
                    )

                    error = BaseExportDataError(
                        message=result.error,
                    )
                    self.result.append_entity(entity=error)

                    break
                else:
                    logger.info(
                        f'Response with {result.response.status_code} code and content {result.response.text}'
                    )

                requests_result.append(result)

    def _prepare(self, *args, **kwargs):
        """
        Выполнение действий функции.
        """
        if self.result.has_not_errors:
            if self.helper.cache.entity_instances:
                self._prepare_data()
                self._prepare_files()
                self._send_files()
            else:
                logger.info(f'{LOGS_DELIMITER * 3} no data for preparing.')

    def before_run(self, *args, **kwargs):
        """
        Действия перед запуском.
        """
        super().before_run(*args, **kwargs)

        logger.info(
            '{delimiter}{force_run}run {runner_name}{log_chunks}..'.format(
                delimiter=LOGS_DELIMITER * 2,
                force_run='force ' if kwargs.get('is_force_run', False) else '',
                runner_name=self.__class__.__name__,
                log_chunks=f' with logs chunk {self._chunk_index}' if self._chunk_index else ''
            )
        )

    def run(self, *args, **kwargs):
        """
        Выполнение действий функции с дальнейшим сохранением объектов в базу при отсутствии ошибок.
        """
        super().run(*args, **kwargs)

        if self.result.has_not_errors:
            self._sub_stage.status_id = ExportingDataSubStageStatus.FINISHED.key
        else:
            self._sub_stage.status_id = ExportingDataSubStageStatus.FAILED.key

        self._sub_stage.save()

        logger.info(f'{LOGS_DELIMITER * 3}change status {repr(self._sub_stage)}')



from itertools import (
    chain,
)
from typing import (
    TYPE_CHECKING,
    Dict,
    Optional,
    Type,
)

import celery
from celery.schedules import (
    crontab,
)
from django.conf import (
    settings,
)
from django.db.models import (
    CharField,
    Max,
)
from django.db.models.functions import (
    Cast,
)
from django.utils import (
    timezone,
)

from educommon.async_task.models import (
    AsyncTaskType,
    RunningTask,
)
from educommon.async_task.tasks import (
    PeriodicAsyncTask,
)
from educommon.utils.date import (
    get_today_min_datetime,
)
from m3_db_utils.consts import (
    DEFAULT_ORDER_NUMBER,
)

from edu_rdm_integration.collect_and_export_data.models import (
    EduRdmCollectDataCommandProgress,
    EduRdmExportDataCommandProgress,
)
from edu_rdm_integration.collect_data.collect import (
    BaseCollectLatestModelsData,
)
from edu_rdm_integration.collect_data.helpers import (
    set_failed_status_suspended_collecting_data_stages,
)
from edu_rdm_integration.consts import (
    REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA,
    REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA,
    TASK_QUEUE_NAME,
)
from edu_rdm_integration.enums import (
    CommandType,
    FileUploadStatusEnum,
)
from edu_rdm_integration.export_data.export import (
    ExportLatestEntitiesData,
)
from edu_rdm_integration.export_data.helpers import (
    set_failed_status_suspended_exporting_data_stages,
)
from edu_rdm_integration.helpers import (
    UploadStatusHelper,
    save_command_log_link,
)
from edu_rdm_integration.models import (
    CollectingDataSubStageStatus,
    CollectingExportedDataSubStage,
    ExportingDataSubStage,
    ExportingDataSubStageStatus,
    ExportingDataSubStageUploaderClientLog,
    RegionalDataMartEntityEnum,
    TransferredEntity,
)
from edu_rdm_integration.storages import (
    RegionalDataMartEntityStorage,
)


if TYPE_CHECKING:
    from datetime import (
        datetime,
    )

    from function_tools.managers import (
        RunnerManager,
    )


class RDMCheckUploadStatus(PeriodicAsyncTask):
    """Периодическая задача для сбора статусов по загрузке файла в витрину."""

    queue = TASK_QUEUE_NAME
    routing_key = TASK_QUEUE_NAME
    description = 'Сбор статусов загрузки данных в витрину "Региональная витрина данных"'
    task_type = AsyncTaskType.UNKNOWN
    run_every = crontab(
        minute=settings.RDM_UPLOAD_STATUS_TASK_MINUTE,
        hour=settings.RDM_UPLOAD_STATUS_TASK_HOUR,
        day_of_week=settings.RDM_UPLOAD_STATUS_TASK_DAY_OF_WEEK,
    )

    def process(self, *args, **kwargs):
        """Выполнение."""
        super().process(*args, **kwargs)

        # Получаем незавершенные загрузки данных в витрину
        in_progress_uploads = ExportingDataSubStageUploaderClientLog.objects.filter(
            file_upload_status=FileUploadStatusEnum.IN_PROGRESS,
            is_emulation=False,
        )

        UploadStatusHelper(in_progress_uploads).run()


class CheckSuspendedExportedStagePeriodicTask(PeriodicAsyncTask):
    """Периодическая задача поиска зависших этапов/подэтапов экспорта."""

    queue = TASK_QUEUE_NAME
    routing_key = TASK_QUEUE_NAME
    description = 'Поиск зависших этапов/подэтапов экспорта в "Региональная витрина данных"'
    task_type = AsyncTaskType.SYSTEM
    run_every = crontab(
        minute=settings.RDM_CHECK_SUSPEND_TASK_MINUTE,
        hour=settings.RDM_CHECK_SUSPEND_TASK_HOUR,
        day_of_week=settings.RDM_CHECK_SUSPEND_TASK_DAY_OF_WEEK,
    )

    def process(self, *args, **kwargs):
        """Выполнение задачи."""
        super().process(*args, **kwargs)

        change_status_collecting_result = set_failed_status_suspended_collecting_data_stages()
        change_status_exporting_result = set_failed_status_suspended_exporting_data_stages()

        task_result = {
            'Прервано сборок': (
                f'Этапов {change_status_collecting_result["change_stage_count"]}'
                f' и подэтапов {change_status_collecting_result["change_sub_stage_count"]}'
            ),
            'Прервано выгрузок': (
                f'Этапов {change_status_exporting_result["change_stage_count"]}'
                f' и подэтапов {change_status_exporting_result["change_sub_stage_count"]}'
            ),
        }

        self.set_progress(
            values=task_result
        )


class TransferLatestEntitiesDataPeriodicTask(PeriodicAsyncTask):
    """Периодическая задача сбора и выгрузки данных."""

    queue = TASK_QUEUE_NAME
    routing_key = TASK_QUEUE_NAME
    description = 'Периодическая задача сбора и экспорта данных РВД'
    task_type = AsyncTaskType.UNKNOWN
    run_every = crontab(
        minute=settings.RDM_TRANSFER_TASK_MINUTE,
        hour=settings.RDM_TRANSFER_TASK_HOUR,
        day_of_week=settings.RDM_TRANSFER_TASK_DAY_OF_WEEK,
    )

    def __init__(self) -> None:
        super().__init__()

        self._collecting_data_managers: Dict[str, Type['RunnerManager']] = {}
        self._collecting_data_manager_to_logs_period_end: Dict[str, datetime] = {}

        self._exporting_data_managers: Dict[str, Type['RunnerManager']] = {}
        self._exporting_data_manager_to_period_end: Dict[str, datetime] = {}

        self._transferred_entities = []

    def process(self, *args, **kwargs):
        """Выполняет задачу."""
        super().process(*args, **kwargs)

        self._collect_transferred_entities()
        self._collect_managers()
        self._calculate_collecting_managers_logs_period_ended_at()
        self._calculate_exporting_managers_ended_at()

        task_id = RunningTask.objects.filter(
            pk=self.request.id,
        ).values_list('pk', flat=True).first()

        model_enum_values = RegionalDataMartEntityEnum.get_entities_model_enums(
            entity_enums=self._transferred_entities,
            is_sorted=True,
        )

        filtered_model_enum_values = filter(lambda model: model.order_number != DEFAULT_ORDER_NUMBER, model_enum_values)
        for model_enum_value in filtered_model_enum_values:
            self._run_collect_model_data(model_enum_value.key, task_id)

        for entity_enum_value in sorted(self._transferred_entities, key=lambda entity: entity.order_number):
            self._run_export_entity_data(entity_enum_value.key, task_id)

    def _run_collect_model_data(self, model: str, task_id: str) -> None:
        """Запускает сбор данных модели РВД."""
        command = self._create_collect_command(model, task_id)
        collect_model_data = self._prepare_collect_model_data_class(command)
        collect_model_data.collect()

        command.refresh_from_db(fields=['stage_id'])
        save_command_log_link(command, settings.RDM_COLLECT_LOG_DIR)

    def _run_export_entity_data(self, entity: str, task_id: str) -> None:
        """Запускает экспорт данных сущности РВД."""
        command = self._create_export_command(entity, task_id)
        if command:
            export_entity_data = self._prepare_export_entity_data_class(command)
            export_entity_data.export()

            command.refresh_from_db(fields=['stage_id'])
            save_command_log_link(command, settings.RDM_EXPORT_LOG_DIR)

    def _collect_transferred_entities(self) -> None:
        """Собирает сущности РВД, по которым будет произведен сбор и экспорт данных."""
        self._transferred_entities = [
            RegionalDataMartEntityEnum.get_model_enum_value(key=entity)
            for entity in TransferredEntity.objects.values_list('entity', flat=True)
        ]

    def _collect_managers(self) -> None:
        """Собирает менеджеры Функций для сбора и выгрузки данных."""
        entity_storage = RegionalDataMartEntityStorage()
        entity_storage.prepare()

        collecting_models_data_managers_map = entity_storage.prepare_entities_manager_map(
            tags={REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA},
        )
        exporting_entities_data_managers_map = entity_storage.prepare_entities_manager_map(
            tags={REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA},
        )

        for transferred_entity in self._transferred_entities:
            for model in chain((transferred_entity.main_model_enum, ), transferred_entity.additional_model_enums):
                collect_manager_class = collecting_models_data_managers_map.get(model.key)
                if collect_manager_class:
                    self._collecting_data_managers[model.key] = collect_manager_class

            export_manager_class = exporting_entities_data_managers_map.get(transferred_entity.key)
            if export_manager_class:
                self._exporting_data_managers[transferred_entity.key] = export_manager_class

    def _calculate_collecting_managers_logs_period_ended_at(self) -> None:
        """Определяет дату последнего успешного этапа сбора у менеджеров Функций сбора."""
        self._collecting_data_manager_to_logs_period_end = dict(
            CollectingExportedDataSubStage.objects.annotate(
                manager_id=Cast('stage__manager_id', output_field=CharField()),
                max_logs_period_ended_at=Max('stage__logs_period_ended_at'),
            ).filter(
                status_id=CollectingDataSubStageStatus.READY_TO_EXPORT.key,
                manager_id__in=[manager.uuid for manager in self._collecting_data_managers.values()],
            ).values_list('manager_id', 'max_logs_period_ended_at'),
        )

    def _calculate_exporting_managers_ended_at(self) -> None:
        """Определяет дату последнего успешного подэтапа экспорта у менеджеров Функций экспорта."""
        self._exporting_data_manager_to_period_end = dict(
            ExportingDataSubStage.objects.annotate(
                manager_id=Cast('stage__manager_id', output_field=CharField()),
                max_ended_at=Max('ended_at'),
            ).filter(
                status_id=ExportingDataSubStageStatus.FINISHED.key,
                manager_id__in=[manager.uuid for manager in self._exporting_data_managers.values()],
            ).values_list('manager_id', 'max_ended_at'),
        )

    def _create_collect_command(self, model: str, task_id: str) -> EduRdmCollectDataCommandProgress:
        """Создает команду сбора данных моделей РВД."""
        manager = self._collecting_data_managers[model]
        logs_period_started_at = (
            self._collecting_data_manager_to_logs_period_end.get(manager.uuid)
            or get_today_min_datetime()
        )

        return EduRdmCollectDataCommandProgress.objects.create(
            model_id=model,
            logs_period_started_at=logs_period_started_at,
            logs_period_ended_at=timezone.now(),
            task_id=task_id,
            type=CommandType.AUTO,
        )

    def _create_export_command(self, entity: str, task_id: str) -> Optional[EduRdmExportDataCommandProgress]:
        """Создает команду экспорта данных сущностей РВД."""
        manager = self._exporting_data_managers[entity]
        period_started_at = self._exporting_data_manager_to_period_end.get(manager.uuid)

        if period_started_at:
            return EduRdmExportDataCommandProgress.objects.create(
                entity_id=entity,
                period_started_at=period_started_at,
                period_ended_at=timezone.now(),
                task_id=task_id,
                type=CommandType.AUTO,
            )

    def _prepare_collect_model_data_class(
        self,
        command: EduRdmCollectDataCommandProgress
    ) -> BaseCollectLatestModelsData:
        """Подготавливает объект класса сбора данных моделей РВД."""
        return BaseCollectLatestModelsData(
            models=[command.model_id],
            logs_period_started_at=command.logs_period_started_at,
            logs_period_ended_at=command.logs_period_ended_at,
            command_id=command.id,
        )

    def _prepare_export_entity_data_class(self, command: EduRdmExportDataCommandProgress) -> ExportLatestEntitiesData:
        """Подготавливает объект класса экспорта данных сущностей РВД.

        При экспорте данных передаем параметр task_id для обновления поля "Описание"
        наименованиями выгруженных сущностей.
        """
        return ExportLatestEntitiesData(
            entities=[command.entity_id],
            period_started_at=command.period_started_at,
            period_ended_at=command.period_ended_at,
            command_id=command.id,
            task_id=self.request.id,
        )


celery_app = celery.app.app_or_default()
celery_app.register_task(RDMCheckUploadStatus)
celery_app.register_task(CheckSuspendedExportedStagePeriodicTask)
celery_app.register_task(TransferLatestEntitiesDataPeriodicTask)
